from transformers import AutoTokenizer, AutoModel
from sentence_transformers import SentenceTransformer
from .llm import LLM_Base
import torch
from torch import Tensor
import torch.nn.functional as F
import pynvml

EMBEDDING_SIZE=1024

class Embedding(LLM_Base):
    embedding_size=EMBEDDING_SIZE
    model_name="intfloat/multilingual-e5-large"
    tokenizer:AutoTokenizer=None
    model:AutoModel=None

    def average_pool(last_hidden_states: Tensor,attention_mask: Tensor) -> Tensor:
        last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
        return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]


    def get_model_name(self):
        return self.model_name
    def set_model_name(self,model_name):
        self.model_name=model_name
    

    def get_available_vram():
        pynvml.nvmlInit()
        handle = pynvml.nvmlDeviceGetHandleByIndex(0)
        mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        return Embedding.bytes_to_gb(mem_info.free)
    def bytes_to_gb(size_in_bytes):
        gb = size_in_bytes / (1024 ** 3)  # 1 GB = 1024^3 bytes
        return gb
    def have_available_gpu():
        if torch.cuda.is_available():
            return Embedding.get_available_vram()>1
        return False
    def get_embeddings(self,sentences:str|list[str]):
        origin_sentences_type=type(sentences)
        if self.model is None:
            self.tokenizer = AutoTokenizer.from_pretrained(self.get_model_name())
            if Embedding.have_available_gpu():
                self.model = AutoModel.from_pretrained(self.get_model_name(),device_map="auto")
            else:
                self.model = AutoModel.from_pretrained(self.get_model_name())
        if origin_sentences_type.__name__=="str":
            sentences=[sentences]
        batch_dict = self.tokenizer(sentences, max_length=512, padding=True, truncation=True, return_tensors='pt')
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') 
        batch_dict.to(device)
        outputs = self.model(**batch_dict)
        embeddings = Embedding.average_pool(outputs.last_hidden_state, batch_dict['attention_mask'])
        tensor_on_cpu = embeddings.cpu()
        if origin_sentences_type.__name__=="str":
            return tensor_on_cpu.detach().numpy()[0]
        return tensor_on_cpu.detach().numpy()
    
    def get_response(self,system,assistant,user):
        return
    def get_response_stream(self,system,assistant,user):
        return
    def get_conversation_stream(self,messages):
        return
    def get_conversation_response(self,messages):
        return
    def detect_if_tokens_oversized(self,e):
        return False
    def get_functions_response(self,messages,functions):
        return


from .llm import LLM
import json

class Formatter:
    def __init__(self, model:LLM,max_repair_count=3):
        self.model=model
        self.max_repair_count=max_repair_count
        self.repair_count=0
    def json(self, json_data, schema=None):
        try:
            json_data=json.loads(json_data)
        except Exception as e:
            print(e)
            json_data=self.repair_json(json_data,schema,e)
        return json_data
    def repair_json(self, json_data, schema=None,error=None):
        system="""Fix the following JSON string to make it valid.
I don't need any extra description in the JSON only give me the JSON.
"""
        if error is not None:
            system+="""Here is the error message during JSON validation:"""+str(error)
        if schema is not None:
            system+="""You can fix the JSON according to the following schema.  Here is the schema:
    """+json.dumps(schema,ensure_ascii=False)
        assistant="""{"""
        user=json_data
        response=self.model.get_response(system,assistant,user)
        self.repair_count+=1
        try:
            json.loads(response)
            return response
        except Exception as e:
            print(e)
            if self.max_repair_count is not None and self.repair_count>=self.max_repair_count:
                raise e
            return self.repair_json(response, schema,e)
    pass


from typing import Callable, Iterator
import time
HEADLESS=True

def execute_queue(iterator:Iterator,threads_count:int=10,slience=True,waiting_time=0):
    def run_task(task:Callable,*args, **kwargs):
        # try:
            task(*args, **kwargs)
        # except Exception as e:
        #     print(e)
    from queue import Queue
    from threading import Thread
    WORKERS_COUNT=threads_count
    queue = Queue()
    def worker():
        continue_running=True
        while continue_running:
            try:
                row=queue.get()
                task=None
                if isinstance(row,tuple) or isinstance(row,list):
                    task=row[0]
                    parameters=row[1]
                if isinstance(row,dict):
                    if "task" in row:task=row["task"]
                    if "function" in row:task=row["function"]
                    if "func" in row:task=row["func"]
                    if "parameter" in row:parameters=row["parameter"]
                    if "parameters" in row:parameters=row["parameters"]
                    if "params" in row:parameters=row["params"]
                    if "args" in row:parameters=row["args"]

                if isinstance(parameters,tuple) or isinstance(parameters,list):
                    args=parameters
                    kwargs={}
                elif isinstance(parameters,dict):
                    args=[]
                    kwargs=parameters
                if not slience:
                    print(f"Running {args}")
                run_task(task,*args, **kwargs)
                if not slience:
                    print(f"Task done.")
                queue.task_done()
            except KeyboardInterrupt:
                print("Main process terminated externally.")
                break
            except BaseException as e:
                continue_running=False
                raise e
    for i in range(WORKERS_COUNT):
        Thread(target=worker, daemon=True).start()
    def start():
        for row in iterator:
            queue.put(row)
            time.sleep(waiting_time)
        queue.join()
    return start()

def build_and_execute_tasks(tasks:list[Callable]|Callable,parameters:Iterator,threads_count:int=10,slience=True,waiting_time=0):
    """Initialize multi-threads queue for parallel tasks
    Args:
        tasks (list[Callable]|Callable): Task that is going to be call.
                            When create your task, please follow the following format.
                            Example:
                                def task(key,value):
                                    return
                                params=[]
                                for key in value:
                                    params.append((key,value))
                                    or
                                    params.append({key=key,value=value})
                                build_and_execute_tasks(task,params)
        parameters (Iterator): Tasks list parameters.
        threads_count (int, optional): How many site you want to scrape at a time. Defaults to 10.
        slience (bool, optional): Stop printing Tasks Running and Done message. Defaults to True.
        waiting_time (int, optional): Wait number of seconds between each threads start. Defaults to 0.
    """
    return build_and_execute(parameters,tasks,threads_count,slience,waiting_time)
def build_and_execute(parameters:Iterator,tasks:list[Callable]|Callable,threads_count:int=10,slience=True,waiting_time=0):
    def run_task(*args, **kwargs):
        try:
            if isinstance(tasks,list):
                for task in tasks:
                    task(*args, **kwargs)
            else:
                tasks(*args, **kwargs)
        except Exception as e:
            print(e)
    from queue import Queue
    from threading import Thread
    WORKERS_COUNT=threads_count
    queue = Queue()
    def worker():
        while True:
            parameter=queue.get()
            if isinstance(parameter,tuple) or isinstance(parameter,list):
                args=parameter
                kwargs={}
            elif isinstance(parameter,dict):
                args=[]
                kwargs=parameter
            if not slience:
                print(f"Running {args}")
            run_task(*args, **kwargs)
            if not slience:
                print(f"Task done.")
            queue.task_done()
    for i in range(WORKERS_COUNT):
        Thread(target=worker, daemon=True).start()
    def start():
        for row in parameters:
            queue.put(row)
            time.sleep(waiting_time)
        queue.join()
    return start()



import os
import json
import datetime
import hashlib
import glob
from abc import ABC, abstractmethod
from typing import Any, Callable, Type,Generator
from .folders import *
import numpy as np

ON_TOKENS_OVERSIZED="on_tokens_oversized"
COMPLETED_STREAM_CACHE={
    "object": "chat.completion.chunk",
    "model": "gpt-4-32k",
    "choices": [
        {
            "index": 0,
            "finish_reason": "stop",
            "delta": {}
        }
    ],
    "usage": None
}

def calculate_md5(string:str):
    md5_hash = hashlib.md5(string.encode()).hexdigest()
    return md5_hash

class CallStack:
    system:str
    assistant:str
    user:str
    response:str
    def __init__(self,system,assistant,user,response) -> None:
        self.system=system
        self.assistant=assistant
        self.user=user
        self.response=response
        pass
    pass
class _LLM_Base(ABC):
    def __init__(self) -> None:
        self.model_name:str=None
        self.use_cache:bool=True
        self.on_each_response:Type[object]=None
        self.on_chunked:Type[object]=None
        pass
    def load_response_cache(model,system,assistant,user):
        try:
            hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
            # print(f"Loading response cache for {model} model with id: {hashed_request}...")
            matching_files = glob.glob(f"{LLM_RESPONSE_CACHE_FOLDER}/{hashed_request}/*.json")
            if len(matching_files)>0:
                with open(matching_files[-1], "r",encoding="utf8") as chat_cache_file:
                    chat_cache = json.load(chat_cache_file)
                    return chat_cache
        except Exception as e:
            print(e)
        return None
    def save_response_cache(model,system,assistant,user,chat_cache):
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        messages=[{"role": "system","content": system},{"role": "user","content": user},{"role": "assistant","content": assistant}]
        _LLM_Base.save_request_log(model,hashed_request,messages,LLM_RESPONSE_CACHE_FOLDER)
        _LLM_Base.save_cache(model,hashed_request,chat_cache,LLM_RESPONSE_CACHE_FOLDER)
    def save_stream_response_cache(model,system,assistant,user,chat_cache,combined=False):
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        messages=[{"role": "system","content": system},{"role": "user","content": user},{"role": "assistant","content": assistant}]
        _LLM_Base.save_request_log(model,hashed_request,messages,LLM_STREAM_RESPONSE_CACHE_FOLDER)
        _LLM_Base.save_cache(model,hashed_request,chat_cache,LLM_STREAM_RESPONSE_CACHE_FOLDER,combined=combined)
    def save_conversation_stream_cache(model,messages,chat_cache):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        _LLM_Base.save_request_log(model,hashed_request,messages,LLM_CONVERSATION_STREAM_CACHE_FOLDER)
        _LLM_Base.save_cache(model,hashed_request,chat_cache,LLM_CONVERSATION_STREAM_CACHE_FOLDER)
    def save_conversation_cache(model,messages,chat_cache):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        _LLM_Base.save_request_log(model,hashed_request,messages,LLM_CONVERSATION_CACHE_FOLDER)
        _LLM_Base.save_cache(model,hashed_request,chat_cache,LLM_CONVERSATION_CACHE_FOLDER)
    def save_cache(model,hashed_request,chat_cache,folder_path,combined=False):
        matching_files = glob.glob(f"{folder_path}/{hashed_request}/*.json")
        file_index=len(matching_files)
        if combined:
            file_index="combined"
        os.makedirs(f"{folder_path}/{hashed_request}", exist_ok=True)
        with open(f"{folder_path}/{hashed_request}/{file_index}.json", "w",encoding="utf8") as temp_file:
            json.dump(chat_cache, temp_file, ensure_ascii=False)
    def save_request_log(model,hashed_request,messages,folder_path):
        request={"messages":messages,"folder_path":folder_path,"hashed_request":hashed_request}
        os.makedirs(f"{folder_path}_requests_logs", exist_ok=True)
        matching_files = glob.glob(f"{folder_path}_requests_logs/{hashed_request}/*.json")
        file_index=len(matching_files)
        os.makedirs(f"{folder_path}_requests_logs/{hashed_request}", exist_ok=True)
        if file_index==0:
            with open(f"{folder_path}_requests_logs/{hashed_request}/{file_index}.json", "w",encoding="utf8") as temp_file:
                json.dump(request, temp_file, ensure_ascii=False)
    def delete_response_cache(model,system,assistant,user):
        _LLM_Base.delete_cache(model,system,assistant,user,LLM_RESPONSE_CACHE_FOLDER)
    def delete_stream_response_cache(model,system,assistant,user):
        _LLM_Base.delete_cache(model,system,assistant,user,LLM_STREAM_RESPONSE_CACHE_FOLDER)
    def delete_cache(model,system,assistant,user,folder_path):
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        print(f"Deleting response cache for {model} model with id: {hashed_request}...")
        matching_files = glob.glob(f"{folder_path}/{hashed_request}/*.json")
        for file in matching_files:
            os.remove(file)
    def split_text_in_half_if_too_large(text:str,max_tokens=10000):
        words = text.split()
        results = []
        
        if len(words) <= max_tokens:
            results.append(' '.join(words))
        else:
            half = len(words) // 2
            results.extend(_LLM_Base.split_text_in_half_if_too_large(' '.join(words[:half])))
            results.extend(_LLM_Base.split_text_in_half_if_too_large(' '.join(words[half:])))
        return results
    def split_text_in_half(text:str):
        words = text.split()
        results = []
        half = len(words) // 2
        results.extend(_LLM_Base.split_text_in_half_if_too_large(' '.join(words[:half])))
        results.extend(_LLM_Base.split_text_in_half_if_too_large(' '.join(words[half:])))
        return results

    # region abstract methods
    @abstractmethod
    def get_model_name(self):
        pass
    @abstractmethod
    def set_model_name(self,name):
        pass
    
    def get_chat_completion(self,*args,**kwargs):
        pass
    
    def detect_if_tokens_oversized(self,e):
        pass
    
    def get_response(self,system,assistant,user)->str:
        pass
    
    def get_response_stream(self,system,assistant,user)->Generator[Any,Any,None]:
        pass
    
    def get_conversation_stream(self,messages)->Generator[Any,Any,None]:
        pass
    
    def get_conversation_response(self,messages)->str:
        pass
    
    def get_embeddings(self,sentences:str|list[str])->np.ndarray:
        pass
    
    # endregion
    def set_event_listener(self,event_name:str,func:Callable[[Any], Any]):
        if event_name=="on_chunked":
            self.on_chunked=func
        elif event_name=="on_each_response":
            self.on_each_response=func
        elif event_name=="on_tokens_oversized":
            self.on_tokens_oversized=func
    
    def is_incomplete_stream_cache(self,chat_cache):
        if "choices" in chat_cache:
            if len(chat_cache["choices"])>0:
                if "finish_reason" in chat_cache["choices"][0]:
                    if chat_cache["choices"][0]["finish_reason"]!="stop":
                        return True
        return False
    
    def on_tokens_oversized(self,e,system,assistant,user):
        if self.detect_if_tokens_oversized(e):
            print("Splitting text in half...")
            chunks = []
            chunks.extend(_LLM_Base.split_text_in_half(user))
            responses=""
            for chunk in chunks:
                try:
                    response=self.get_response(system,assistant,chunk)
                except Exception as e:
                    print(e)
                    continue
                if response is not None:
                    if self.on_chunked is None:
                        responses+=response
                    else:
                        responses=self.on_chunked(system,assistant,chunk,responses,response)
            return responses
    
    def have_stream_response_cache(self,model,system,assistant,user):
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        return os.path.exists(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}")
    def load_stream_response_cache(self,model,system,assistant,user):
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        # print(f"Loading response cache for {model} model with id: {hashed_request}...")
        if self.have_stream_response_cache(model,system,assistant,user):
            if os.path.exists(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}/combined.json"):
                with open(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}/combined.json", "r",encoding="utf8") as chat_cache_file:
                    chat_cache = json.load(chat_cache_file)
                    for chunk in chat_cache:
                        yield chunk
                    return
            matching_files = glob.glob(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}/*.json")
            matching_files=sorted(matching_files, key=lambda x: int(os.path.basename(x).split(".")[0]))
            for path in matching_files:
                with open(path, "r",encoding="utf8") as chat_cache_file:
                    chat_cache = json.load(chat_cache_file)
                    yield chat_cache
            if self.is_incomplete_stream_cache(chat_cache):
                chat_cache=COMPLETED_STREAM_CACHE
                yield chat_cache
                import shutil
                shutil.rmtree(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}")
        return
    
    def combine_stream_response_cache(self,system,assistant,user):
        model=self.get_model_name()
        hashed_request=calculate_md5(f"{model}{system}{assistant}{user}")
        if os.path.exists(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}/combined.json"):
            return
        if self.have_stream_response_cache(model,system,assistant,user):
            matching_files = glob.glob(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}/*.json")
            matching_files=sorted(matching_files, key=lambda x: int(os.path.basename(x).split(".")[0]))
            chat_caches=[]
            for path in matching_files:
                with open(path, "r",encoding="utf8") as chat_cache_file:
                    chat_cache = json.load(chat_cache_file)
                    chat_caches.append(chat_cache)
            if self.is_incomplete_stream_cache(chat_cache) is False:
                import shutil
                shutil.rmtree(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}")
                _LLM_Base.save_stream_response_cache(model,system,assistant,user,chat_caches,combined=True)

    
    def have_conversation_stream_cache(self,model,messages:list):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        return os.path.exists(f"{LLM_STREAM_RESPONSE_CACHE_FOLDER}/{hashed_request}")
    def load_conversation_stream_cache(self,model,messages:list):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        # print(f"Loading response cache for {model} model with id: {hashed_request}...")
        if self.have_conversation_stream_cache(model,messages):
            matching_files = glob.glob(f"{LLM_CONVERSATION_STREAM_CACHE_FOLDER}/{hashed_request}/*.json")
            matching_files=sorted(matching_files, key=lambda x: int(os.path.basename(x).split(".")[0]))
            for path in matching_files:
                with open(path, "r",encoding="utf8") as chat_cache_file:
                    chat_cache = json.load(chat_cache_file)
                    yield chat_cache
            if self.is_incomplete_stream_cache(chat_cache):
                chat_cache=COMPLETED_STREAM_CACHE
                yield chat_cache
                import shutil
                shutil.rmtree(f"{LLM_CONVERSATION_STREAM_CACHE_FOLDER}/{hashed_request}")
    def have_conversation_cache(self,model,messages:list):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        return os.path.exists(f"{LLM_CONVERSATION_CACHE_FOLDER}/{hashed_request}.json")
    def load_conversation_cache(self,model,messages:list):
        hashed_request=calculate_md5(f"{model}{json.dumps(messages)}")
        # print(f"Loading response cache for {model} model with id: {hashed_request}...")
        matching_files = glob.glob(f"{LLM_CONVERSATION_CACHE_FOLDER}/{hashed_request}/*.json")
        if len(matching_files)>0:
            with open(matching_files[-1], "r",encoding="utf8") as chat_cache_file:
                chat_cache = json.load(chat_cache_file)
                return chat_cache
        return None
    
    def get_request_hash(self,*args,**kwargs):
        params=[]
        for arg in args:
            if isinstance(arg,dict):
                params.append(self.get_request_hash(**arg))
            if isinstance(arg,list):
                params.append(self.get_request_hash(*arg))
            if isinstance(arg,str) or isinstance(arg,int) or isinstance(arg,float):
                params.append(str(arg))
            pass
        for key in kwargs:
            if isinstance(kwargs[key],dict):
                params.append(self.get_request_hash(**kwargs[key]))
            if isinstance(kwargs[key],list):
                params.append(self.get_request_hash(*kwargs[key]))
            if isinstance(kwargs[key],str) or isinstance(kwargs[key],int) or isinstance(kwargs[key],float):
                params.append(str(kwargs[key]))
            pass
        return calculate_md5("".join(params))
    def have_chat_completion_cache(self,hashed_request=None,*args,**kwargs):
        if hashed_request is None:
            hashed_request=self.get_request_hash(*args,**kwargs)
        return os.path.exists(f"{LLM_CHAT_COMPLETION_FOLDER}/{hashed_request}")
    def have_chat_completion_cache_stream(self,hashed_request=None,*args,**kwargs):
        if hashed_request is None:
            hashed_request=self.get_request_hash(*args,**kwargs)
        return os.path.exists(f"{LLM_CHAT_COMPLETION_STREAM_FOLDER}/{hashed_request}")

    def save_chat_completion_questions(self,*args,**kwargs):
        LLM_Base.save_cache(*args,**kwargs,folder_path=LLM_CHAT_COMPLETION_QUESTIONS_FOLDER)
    def save_chat_completion_cache_stream(self,*args,**kwargs):
        LLM_Base.save_cache(*args,**kwargs,folder_path=LLM_CHAT_COMPLETION_STREAM_FOLDER)
    def save_chat_completion_cache(self,*args,**kwargs):
        LLM_Base.save_cache(*args,**kwargs,folder_path=LLM_CHAT_COMPLETION_FOLDER)
    def load_chat_completion_cache_stream(self,hashed_request=None,*args,**kwargs):
        if hashed_request is None:
            model=self.get_model_name()
            hashed_request=self.get_request_hash(model,*args,**kwargs)
        if self.have_chat_completion_cache_stream(hashed_request,*args,**kwargs):
            matching_files = glob.glob(f"{LLM_CHAT_COMPLETION_STREAM_FOLDER}/{hashed_request}/*.json")
            matching_files=sorted(matching_files, key=lambda x: int(os.path.basename(x).split(".")[0]))
            for path in matching_files:
                with open(path, "r",encoding="utf8") as chat_caches_file:
                    chat_caches = json.load(chat_caches_file)
                    for chat_cache in chat_caches:
                        yield chat_cache
        else:
            return None
    def load_chat_completion_cache(self,hashed_request=None,*args,**kwargs):
        if hashed_request is None:
            model=self.get_model_name()
            hashed_request=self.get_request_hash(model,*args,**kwargs)
        if self.have_chat_completion_cache(hashed_request,*args,**kwargs):
            matching_files = glob.glob(f"{LLM_CHAT_COMPLETION_FOLDER}/{hashed_request}/*.json")
            matching_files=sorted(matching_files, key=lambda x: int(os.path.basename(x).split(".")[0]))
            for path in matching_files:
                with open(path, "r",encoding="utf8") as chat_caches_file:
                    chat_caches = json.load(chat_caches_file)
                    return chat_caches
        else:
            return None
    pass


class LLM_Base(_LLM_Base):
    embedding_size=None
    def __init__(self,instant:_LLM_Base) -> None:
        self.instant=instant
        pass
    pass
class LLM:
    class DictWrapper:
        def __init__(self, dictionary):
            self.__dict__ = {}
            for key, value in dictionary.items():
                if isinstance(value, dict):
                    self.__dict__[key] = LLM.DictWrapper(value)
                elif isinstance(value, list):
                    self.__dict__[key] = LLM.ListWrapper(value)
                else:
                    self.__dict__[key] = value

        def __getattr__(self, key):
            try:
                return self.__dict__[key]
            except KeyError:
                raise AttributeError(key)

        def __setattr__(self, key, value):
            self.__dict__[key] = value
        def to_dict(self):
            result = {}
            for key, value in self.__dict__.items():
                if isinstance(value, LLM.DictWrapper):
                    result[key] = value.to_dict()
                elif isinstance(value, LLM.ListWrapper):
                    result[key] = value.to_list()
                else:
                    result[key] = value
            return result
        pass
    class ListWrapper:
        def __init__(self, lst):
            self.__list__ = []
            for item in lst:
                if isinstance(item, dict):
                    self.__list__.append(LLM.DictWrapper(item))
                elif isinstance(item, list):
                    self.__list__.append(LLM.ListWrapper(item))
                else:
                    self.__list__.append(item)

        def __getitem__(self, index):
            return self.__list__[index]

        def __len__(self):
            return len(self.__list__)
        def to_list(self):
            result = []
            for item in self.__list__:
                if isinstance(item, LLM.DictWrapper):
                    result.append(item.to_dict())
                elif isinstance(item, LLM.ListWrapper):
                    result.append(item.to_list())
                else:
                    result.append(item)
            return result
        pass
    
    def __init__(self,ModelClass:Type[LLM_Base],use_cache:bool=True,
                 on_each_response:Callable[[str,str,str,str,str], str]=None,
                 on_chunked:Callable[[str,str,str,str,str], str]=None) -> None:
        """
        Constructor for LLM class
        :param ModelClass: LLM_Base The class of the model to be used
        :param use_cache: bool Whether to use cache or not
        :param on_each_response: Callable[[str,str,str,str,str], str] A function to be called on each response
            - This function should take 5 string from get_response or in between on_tokens_oversized: 
                system,assistant,user,responses,response, and the return value should be a string
            - for get_response, the responses is None and the response is the current response.
            - for on_tokens_oversized, the responses is the concatenated string of previous responses and the response is the current response.
                because the input is too large, on tokens oversized is called recursively, 
                it has become complicated to get back the concatenated responses.
            - for system,assistant,user, it will send the string needed to generate current response
        """
        self.model_class=ModelClass(self)
        self.use_cache=use_cache
        self.on_each_response=on_each_response
        self.on_chunked=on_chunked

        self.model_class.use_cache=self.use_cache
        self.model_class.on_each_response=self.on_each_response
        self.model_class.on_chunked=self.on_chunked
        pass
    def get_model_name(self):
        return self.model_class.get_model_name()
    def set_model_name(self,name):
        return self.model_class.set_model_name(name)
    def get_chat_completion(self,*args,**kwargs):
        return self.model_class.get_chat_completion(*args,**kwargs)
    def get_response(self,system,assistant,user)->str:
        if self.on_each_response is None:
            return self.model_class.get_response(system,assistant,user)
        else:
            return self.on_each_response(system,assistant,user,None,self.model_class.get_response(system,assistant,user))
    def get_json_response(self,system,assistant,user)->dict:
        response=self.get_response(system,user,assistant)
        try:
            return json.loads(response)
        except Exception as e:
            print(e)
            print(response)
            return self.repair_json(response)
    def repair_json(self,json_data, schema=None,error=None)->dict:
        system="""Fix the following JSON string to make it valid.
I don't need any extra description in the JSON only give me the JSON.
"""
        if error is not None:
            system+="""Here is the error message during JSON validation:"""+str(error)
        if schema is not None:
            system+="""You can fix the JSON according to the following schema.  Here is the schema:
"""+json.dumps(schema,ensure_ascii=False)
        assistant="""{"""
        user=json_data
        response=self.get_response(system,assistant,user)
        try:
            return json.loads(response)
        except Exception as e:
            print(e)
            print(response)
            return self.repair_json(response, schema,e)
    
    def get_response_stream(self,system,assistant,user):
        response=self.model_class.get_response_stream(system,assistant,user)
        return response
    def combine_stream_response_cache(self,system,assistant,user):
        return self.model_class.combine_stream_response_cache(system,assistant,user)
    def get_conversation_stream(self,messages):
        return self.model_class.get_conversation_stream(messages)
    def on_tokens_oversized(self,e,system,assistant,user):
        return self.model_class.on_tokens_oversized(e,system,assistant,user)
    def set_event_listener(self,event_name:str,func:Callable[[Any], Any]):
        self.model_class.set_event_listener(event_name,func)
    def get_conversation_response(self,messages):
        return self.model_class.get_conversation_response(messages)
    def get_embeddings(self,sentences:str|list[str])->np.ndarray:
        return self.model_class.get_embeddings(sentences)
    
    pass


import os
def reset_root_folder(path:str):
    global ROOT_FOLDER,\
        LLM_CACHES_FOLDER,\
        LLM_RESPONSE_CACHE_FOLDER,\
        LLM_STREAM_RESPONSE_CACHE_FOLDER,\
        LLM_CONVERSATION_STREAM_CACHE_FOLDER,\
        LLM_CONVERSATION_CACHE_FOLDER,\
        LLM_CHAT_COMPLETION_QUESTIONS_FOLDER,\
        LLM_CHAT_COMPLETION_FOLDER,\
        LLM_CHAT_COMPLETION_STREAM_FOLDER,\
        LLM_MODELS_FOLDER
    ROOT_FOLDER=path
    LLM_CACHES_FOLDER = os.path.join(ROOT_FOLDER,"llm_caches")
    LLM_RESPONSE_CACHE_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_response_cache")
    LLM_STREAM_RESPONSE_CACHE_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_stream_response_cache")
    LLM_CONVERSATION_STREAM_CACHE_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_conversation_stream_cache")
    LLM_CONVERSATION_CACHE_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_conversation_cache")
    LLM_CHAT_COMPLETION_QUESTIONS_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_chat_completion_questions")
    LLM_CHAT_COMPLETION_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_chat_completion_chunk")
    LLM_CHAT_COMPLETION_STREAM_FOLDER = os.path.join(LLM_CACHES_FOLDER,"llm_chat_completion_stream")
    LLM_MODELS_FOLDER = os.path.join(ROOT_FOLDER,"llm_models")

reset_root_folder(os.getcwd())